
;
;       Bang! - A Unix-like history for DOS
;       Copyright 1990 by Lawrence Kesteloot
;       Version 1.0 - August 26, 1990
;
;       Assemble with Turbo-Assembler to create .com file:
;
;               tasm bang!
;               tlink -t bang!
;
;       First written: August 24, 1990
;       Last update:   August 26, 1990
;

MAXHIST EQU     20d             ; 20 entries in history max
CR      EQU     13d             ; Carriage return
LF      EQU     10d             ; Line feed

codseg  segment public  'code'
        assume  cs:codseg
        assume  ds:codseg
        assume  ss:codseg
        assume  es:codseg

        org     100h

start:
        mov     sp,offset stktop
        mov     bx,cs
        mov     ds,bx

        mov     word ptr [cs:promptnum],1 ; TRUE - do print hist num
        xor     ch,ch
        mov     cl,[es:80h]             ; Command-line parameters length
        inc     cx                      ; Count C/R also
        mov     di,81h                  ; Point to command-line
nextcmdlinechar:
        mov     al,[es:di]              ; Get character
        cmp     al,'p'                  ; Check if "/P" for prompt hist num
        je      doprintnum
        cmp     al,'P'
        je      doprintnum
        inc     di
        loop    nextcmdlinechar
        mov     word ptr [cs:promptnum],0 ; FALSE - don't print hist num
doprintnum:

        mov     ah,40h
        mov     bx,1
        mov     dx,offset initstr
        mov     cx,strlen
        int     21h

        mov     ax,0                    ; clear buffer
        mov     cs:[histsize],ax

        cli
        mov     ah,35h
        mov     al,21h
        int     21h
        mov     word ptr cs:oldint,bx
        mov     word ptr cs:oldint+2,es

        cli
        mov     dx,offset newint
        mov     ah,25h
        mov     al,21h
        int     21h

        sti
        mov     al,0
        mov     ah,31h
        mov     dx,offset progend
        mov     cl,4
        shr     dx,cl
        add     dx,2
        int     21h

; ***************************************************************************

newint  proc    far

        sti
        cmp     ah,0Ah                  ; Skip everything not GETSTRING
        jne     skip

        pushf                           ; Simulate an INT
        push    cs
        mov     cs:[oldax],ax
        mov     ax,offset routine       ; Catch routine on way back
        push    ax
        mov     ax,cs:[oldax]

        cmp     word ptr [cs:promptnum],0 ; Check to see to print num
        je      skip

        push    bx
        push    ax                      ; Display history number on
        mov     ax,[cs:histsize]        ; prompt line
        inc     ax                      ; Display NEXT history number
        mov     bx,1                    ; Don't print spaces
        call    dispnum
        pop     ax
        pop     bx

skip:
        jmp     dword ptr cs:oldint     ; Call INT 21h anyway

; ***************************************************************************

routine:
        push    ds
        push    es
        push    si
        push    di
        push    bp
        push    ax
        push    bx
        push    cx
        push    dx

        mov     bp,sp                   ; BP always points to the original
                                        ; DX, which points to the input
                                        ; buffer.

beginning:
        mov     bx,dx

        mov     al,ds:[bx+1]            ; Check length of string
        mov     cl,al                   ; Save for later
        cmp     al,0
        je      endint                  ; If nothing, then skip routine

        mov     al,ds:[bx+2]            ; If first character is not !,
        cmp     al,'!'                  ; then save current line
        je      @@1
        jmp     saveline
@@1:
        cmp     cl,1                    ; If just "!" then display history
        je      disphist

        mov     al,ds:[bx+3]            ; Check 2nd character
        cmp     al,'!'                  ; If it's "!xxx" then recall old
        je      @@2                     ; entry
        jmp     recallold
@@2:
        mov     al,ds:[bx+4]            ; If "!!!"
        cmp     al,'!'
        je      @@3
        jmp     recalllast
@@3:
        push    bx
        push    ds
        mov     ax,cs
        mov     ds,ax
        mov     ah,40h
        mov     bx,1
        mov     dx,offset instru
        mov     cx,instrlen
        int     21h
        pop     ds
        pop     bx

clearline:
        mov     bx,[bp]                 ; Make sure BX points to buffer
        mov     byte ptr [ds:bx+2],CR   ; Set line to null
        mov     byte ptr [ds:bx+1],0    ; Set length to zero

endint:
        pop     dx
        pop     cx
        pop     bx
        pop     ax
        pop     bp
        pop     di
        pop     si
        pop     es
        pop     ds
        iret

; ***************************************************************************

disphist:                               ; Display history
        push    ds
        mov     ax,cs
        mov     ds,ax

        mov     ah,2                    ; Display character
        mov     dl,LF
        int     21h

        mov     ax,cs:[histsize]        ; Get length
        sub     ax,MAXHIST              ; Get top of history
        jge     again                   ; If fewer than MAXHIST items in hist
        xor     ax,ax                   ; then display only those
again:
        inc     ax
        cmp     ax,cs:[histsize]        ; Check if done
        jg      donedisp
        push    ax                      ; Save for later
        mov     bx,0                    ; Print leading spaces
        call    dispnum                 ; Display the number AX

        pop     ax
        push    ax

        call    getptr                  ; Get pointer into history
        mov     si,ax                   ; Move into source index

        xor     cx,cx
        mov     cl,[si]                 ; Put length into cx
        inc     si                      ; Skip length
        mov     dx,si                   ; ds:dx --> string
        mov     ah,40h                  ; Display string
        mov     bx,1                    ; to screen
        int     21h

        mov     ah,2                    ; Display character
        mov     dl,CR
        int     21h
        mov     ah,2                    ; Display character
        mov     dl,LF
        int     21h

        pop     ax
        jmp     again

donedisp:
        pop     ds
        jmp     clearline

; ***************************************************************************
;   AL already holds the second character, BX points to buffer
 
recallold:
        cmp     al,'0'
        jl      matchstr
        cmp     al,'9'
        jg      matchstr

        xor     ax,ax                   ; ax will keep number
        mov     si,bx                   ; keep index in si
        add     si,3                    ; start at 2nd character
        xor     cx,cx                   ; clear to start
getnext:
        mov     cl,[si]                 ; Get digit in cl
        cmp     cl,'0'                  ; Check to see if end of number
        jl      donenum
        cmp     cl,'9'
        jg      donenum
        sub     cl,'0'                  ; convert to normal byte
        mov     bx,10                   ; Convert from base 10
        mul     bx                      ; dx:ax <-- ax*10
        add     ax,cx                   ; add current digit
        inc     si
        jmp     getnext
donenum:
        mov     bx,cs:[histsize]
        cmp     ax,bx                   ; see if out of range
        jg      rangeerror              ; too large
        sub     bx,MAXHIST-1
        cmp     ax,bx
        jl      rangeerror              ; too small
        cmp     ax,0
        je      rangeerror              ; never a zero

        jmp     copyline

rangeerror:
        mov     bx,[bp]                 ; displayerror wants BX --> buffer
        call    displayerror

        jmp     clearline

; *********** MATCH STRING:

matchstr:
        push    es
        mov     ax,cs                   ; For ES:DI
        mov     es,ax

        mov     si,bx                   ; BX is pointer to buffer
        add     si,3                    ; Point to 2nd character
        push    si
        xor     dx,dx                   ; DX will store length of string
nextchar:
        inc     dx
        cmp     byte ptr [si],CR        ; See if at end of string
        je      endofstring
        cmp     byte ptr [si],' '       ; See if at end of word
        je      endofstring
        inc     si
        jmp     nextchar
endofstring:
        pop     si                      ; Point back to 2nd character
        mov     ax,cs:[histsize]        ; Start at last command
        cld                             ; Go forward
previousline:
        push    si
        push    ax                      ; Save for later
        call    getptr                  ; Get pointer into history
        mov     di,ax                   ; And put into destination (of compare)
        inc     di                      ; Point to first character (skip len)

        mov     cx,dx                   ; CX stores length of string
        repe    cmpsb                   ; Find first mismatch

        pop     ax
        pop     si                      ; Make SI point to buffer again

        jcxz    donesearching           ; Strings were identical

        dec     ax                      ; Get command before that
        mov     bx,cs:[histsize]        ; Find if done
        sub     bx,MAXHIST-1
        cmp     ax,bx
        jl      matchnotfound
        cmp     ax,1
        jl      matchnotfound

        jmp     previousline

donesearching:
        pop     es
        jmp     copyline

matchnotfound:
        pop     es
        mov     bx,[bp]                 ; displayerror wants BX --> buffer
        call    displayerror

        jmp     clearline

; ***************************************************************************

recalllast:
        mov     ax,cs:[histsize]
        cmp     ax,0
        jne     @@4
        jmp     clearline
@@4:
        jmp     copyline


; ***************************************************************************
;   Copies the event number AX into the buffer pointed to by the value
;   pointed to by BP

copyline:
        push    ds

; First check to see if there is an addition to the recalled cmd
; e.g.,  !d /w

        mov     bx,[bp]                 ; Load BX with pointer to buffer
        add     bx,4                    ; Point to 3rd character
        mov     byte ptr cs:[tempbuf],CR ; Clear temporary buffer

checknextchar:
        cmp     byte ptr ds:[bx],CR     ; See if end of string
        je      noaddition
        cmp     byte ptr ds:[bx],' '    ; See if end of keystring
        je      addition
        inc     bx
        jmp     checknextchar

addition:
        mov     di,offset tempbuf
copynextchar:
        mov     dl,[bx]                 ; Copy char to temp buffer
        mov     cs:[di],dl

        cmp     dl,CR                   ; Check if end of string
        je      noaddition
        inc     bx
        inc     di
        jmp     copynextchar

noaddition:
        call    getptr                  ; Get pointer into history
        mov     si,ax                   ; Move into source index

        mov     ax,ds                   ; es:di points to buffer
        mov     es,ax

        mov     ax,cs                   ; ds:si points to history
        mov     ds,ax

        mov     bx,[bp]                 ; BX will be used later
        mov     di,bx                   ; es:di points to buffer (dest)
        inc     di                      ; Ignore first byte (max length)

        mov     cx,64                   ; Length of string to copy in words

        cld                             ; Clear Direction bit - go forward
        rep     movsw                   ; Copy to history

        mov     ah,2                    ; Display character (LF)
        mov     dl,LF
        int     21h

        pop     ds

        mov     cl,[bx+1]               ; Load official length into CX
        xor     ch,ch
        mov     si,cx
        lea     di,[bx+si+2]            ; Point to end of buffer
        mov     si,offset tempbuf       ; Now copy any addition to buffer
nextadd:
        mov     dl,cs:[si]
        mov     es:[di],dl
        cmp     dl,CR
        je      donecopy
        inc     si
        inc     di
        jmp     nextadd
donecopy:
        sub     di,bx                   ; Find real length (after add)
        sub     di,2                    ; Subtract CR and Length
        mov     cx,di                   ; Put into length CX
        mov     [bx+1],cl               ; And store permanantly

        mov     ah,40h                  ; Print string
        lea     dx,[bx+2]               ; ds:dx points to string
        mov     bx,1                    ; To screen
        int     21h

        mov     ah,2                    ; Display character
        mov     dl,CR
        int     21h

        mov     dx,[bp]                 ; get dx ready for "beginning"

        jmp     beginning               ; Parse line again

; ***************************************************************************

saveline:
        inc     cs:[histsize]           ; Next line
        mov     ax,cs:[histsize]
        call    getptr                  ; Find pointer into history
        mov     di,ax                   ; Move into destination index

        mov     ax,cs                   ; es:di points to history
        mov     es,ax

        mov     si,bx                   ; ds:si points to buffer (source)
        inc     si                      ; Ignore first byte (max length)

        mov     cx,64                   ; Length of string to copy in words

        cld                             ; Clear Direction bit - go forward
        rep     movsw                   ; Copy to history

        jmp     endint

newint  endp

; ***************************************************************************
;   Given AX, returns AX which points to offset in history of entry
;   Destroys nothing except AX

getptr  proc    near

        push    dx
        push    cx
        mov     dl,MAXHIST              ; Divide by MAXHIST to get index
        div     dl                      ; Remainder in ah
        xor     al,al                   ; Clear low byte
        xchg    al,ah                   ; Swap to get AX straight
        mov     cl,7                    ; 2^7=128
        shl     ax,cl                   ; Multiply by 128 to get pointer
        add     ax,offset history       ; Move to array
        pop     cx
        pop     dx

        ret

getptr  endp

; ***************************************************************************
; If BX=0:
;   Displays AX in base 10 in exactly 6 characters with leading spaces
;   and two spaces afterwards (e.g., "    12  ")
; If BX=1:
;   Displays AX in base 10 in as many digits as it needs, with no
;   leading or trailing spaces, and a ">" afterwards (e.g., "12>")
; Destroys nothing

dispnum proc    near

        push    ax
        push    bx
        push    cx
        push    dx
        push    si

        mov     cx,bx                   ; Put flag in CX

        cmp     cx,1                    ; Don't print leading space
        je      @@10

        push    ax
        mov     ah,2                    ; one space to begin with
        mov     dl,' '
        int     21h
        pop     ax
@@10:
        xor     bx,bx                   ; No characters printed yet
        xor     si,si                   ; Point to first divisor (10000)

nextdig:
        xor     dx,dx                   ; Divide dx:ax by divarr[si]
        div     cs:[divarr+si]          ; Result in ax, remainder in dx
        add     si,2                    ; Point to next divisor

        push    dx

        mov     ah,2                    ; Display character
        add     al,'0'                  ; Make it printable
        cmp     al,'0'                  ; If it is not a zero
        jne     printdigit              ; Then print no matter what
        cmp     bx,0                    ; See if anything has been printed yet
        jne     printdigit              ; If so, then print zeros
        cmp     si,10                   ; If last character
        je      printdigit              ; Then print zero anyway
        mov     al,' '                  ; print space if leading char
        cmp     cx,1                    ; Don't print leading space
        je      @@20
        jmp     printedspace            ; Don't set BX
printdigit:
        mov     bx,1                    ; Set PRINTED flag to true
printedspace:
        mov     dl,al                   ; Put character into dl
        int     21h                     ; Print the sucker
@@20:
        pop     ax                      ; Put remainder back into ax

        cmp     si,10                   ; See if done
        jne     nextdig

        cmp     cx,1                    ; Don't print trailing spaces
        je      @@30

        mov     ah,2                    ; Display character
        mov     dl,' '                  ; Print space
        int     21h
        mov     ah,2                    ; Display character
        mov     dl,' '                  ; Print space
        int     21h
        jmp     @@40
@@30:
        mov     ah,2                    ; Display character
        mov     dl,'>'                  ; Print prompt
        int     21h
@@40:
        pop     si
        pop     dx
        pop     cx
        pop     bx
        pop     ax

        ret

dispnum endp

; ***************************************************************************
;   Displays the input string (without the !) and displays the error
;   message.  BX should point to input buffer in DS.  Then returns.

displayerror    proc       near

        mov     ah,2                    ; Display character
        mov     dl,LF
        int     21h

        mov     ah,40h                  ; print string
        mov     cl,[bx+1]               ; length of string
        xor     ch,ch                   ; move cl into cx
        dec     cx                      ; Don't include "!"
        lea     dx,[bx+3]               ; ds:dx points to 2nd char
        mov     bx,1                    ; to screen
        int     21h

        push    ds
        mov     ax,cs
        mov     ds,ax
        mov     ah,40h                  ; print string
        mov     bx,1                    ; to screen
        mov     cx,lenstrerror          ; length of string
        mov     dx,offset strerror      ; ds:dx points to string
        int     21h
        pop     ds

        ret

displayerror    endp

; ***************************************************************************

initstr db      '"Bang!" now installed - Type "!!!" for instructions',13,10
strlen  equ     $ - initstr

instru  db      10,10,'Bang! - A Unix-like history for DOS - Version 1.0 - August 26, 1990',13,10
        db      'Copyright 1990 by Lawrence Kesteloot',13,10,10
        db      'To display the history list, type "!"',13,10,10
        db      'To recall the last command, type "!!"',13,10,10
        db      'To recall a previous command, type "!str" where "str" is the beginning of the',13,10
        db      'string to match.  For example, if "DIR /W" was a previous command, then "!D"',13,10
        db      'or "!DI" would recall it.  Note that the search is case-sensitive.  This allows',13,10
        db      'you to do things like switch between two directories, "CD DIR1" and "cd dir2"',13,10
        db      'by typing "!C" and "!c".',13,10,10
        db      'Also, an old command may be invoked by typing "!n" where "n" is the number',13,10
        db      'displayed next to the line in the history list.',13,10,10
        db      'Parameters may even be added, as in "!D /P" would give "DIR /W /P".',13,10,10
        db      'If Bang! is invoked with the "/P" switch, then the history number will be added',13,10
        db      'to the prompt.',13,10
        
instrlen equ    $ - instru

strerror db     ': Event not found.',13,10
lenstrerror equ $ - strerror

divarr  dw      10000,1000,100,10,1     ; For HEX --> DEC routine
promptnum dw    ?                       ; 1 if printing history num on prompt
oldint  dd      ?                       ; Old INT 21h
oldax   dw      ?                       ; Temp space for AX
histsize dw     ?                       ; Largest history number so far
history db      128*MAXHIST dup (?)     ; Space to actually holdhistory
tempbuf db      128 dup (?)             ; Temp buffer for anything
stkbot  db      200 dup (?)             ; Temp stack.  Not really used much
stktop  db      ?
progend db      ?                       ; To determine size of program

codseg  ends

        end     start

